/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, FRAUNHOFER HHI
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of FRAUNHOFER HHI
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/


/** \file     TComPredFilterMOMS.cpp
    \brief    MOMS interpolation filter
*/

#include "TComPredFilterMOMS.h"

#if HHI_INTERP_FILTER

// predict luma block for a given motion vector for motion compensation
Void  TComPredFilterMOMS::predInterLumaBlkMOMS( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv, Int iWidth, Int iHeight, TComYuv*& rpcYuv, InterpFilterType ePFilt, Int* piExtY )
{
  Int     iRefStride  = pcPicYuvRef->getStride();
  Int     iDstStride  = rpcYuv->getStride();

  Int     iRefOffset  = ( pcMv->getHor() >> 2 ) + ( pcMv->getVer() >> 2 ) * iRefStride;
  Pel*    piRefY      = pcPicYuvRef->getLumaAddr( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr ) + iRefOffset;

  Int     ixFrac      = pcMv->getHor() & 0x3;
  Int     iyFrac      = pcMv->getVer() & 0x3;

  Pel*    piDstY      = rpcYuv->getLumaAddr( uiPartAddr );

  setFiltType( ePFilt );
  m_pcInterpIf->interpolate( piDstY, iDstStride, piRefY, iRefStride, iHeight, iWidth, ixFrac<<1, iyFrac<<1, 1, piExtY );
}


// predict chroma block for a given motion vector for motion compensation
Void TComPredFilterMOMS::predInterChromaBlkMOMS( TComDataCU* pcCU, TComPicYuv* pcPicYuvRef, UInt uiPartAddr, TComMv* pcMv, Int iWidth, Int iHeight, TComYuv*& rpcYuv, InterpFilterType ePFilt )
{
  Int     iRefStride  = pcPicYuvRef->getCStride();
  Int     iDstStride  = rpcYuv->getCStride();

  Int     iRefOffset  = (pcMv->getHor() >> 3) + (pcMv->getVer() >> 3) * iRefStride;

  Pel*    piRefCb     = pcPicYuvRef->getCbAddr( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr ) + iRefOffset;
  Pel*    piRefCr     = pcPicYuvRef->getCrAddr( pcCU->getAddr(), pcCU->getZorderIdxInCU() + uiPartAddr ) + iRefOffset;

  Pel*    piDstCb     = rpcYuv->getCbAddr( uiPartAddr );
  Pel*    piDstCr     = rpcYuv->getCrAddr( uiPartAddr );

  Int     ixFrac      = pcMv->getHor() & 0x7;
  Int     iyFrac      = pcMv->getVer() & 0x7;

  setFiltType( ePFilt );
  m_pcInterpIf->interpolate( piDstCb, iDstStride, piRefCb, iRefStride, iHeight>>1, iWidth>>1, ixFrac, iyFrac, 1 );
  m_pcInterpIf->interpolate( piDstCr, iDstStride, piRefCr, iRefStride, iHeight>>1, iWidth>>1, ixFrac, iyFrac, 1 );
}


// half-pixel upsampling for motion estimation
Void TComPredFilterMOMS::extMOMSUpSamplingH ( TComPattern* pcPattern )
{
  Int   iDx, iDy;

  Int   iWidth      = pcPattern->getROIYWidth();
  Int   iHeight     = pcPattern->getROIYHeight();

  Int   iSrcStride  = pcPattern->getPatternLStride();
  Int   iBufStride  = m_cYuvTmp.getStride();

  Pel*  piDstYPel;
  Pel*  piSrcYPel;

  piSrcYPel = pcPattern->getROIY() - iSrcStride - 1;
  piDstYPel = m_cYuvTmp.getLumaAddr();
  for ( iDy = 0; iDy <= 2; iDy += 2 )
  {
    for ( iDx = 0; iDx <= 2; iDx += 2 )
    {
      m_pcInterpIf->interpolate( piDstYPel+(iDx+iDy*iBufStride), iBufStride, piSrcYPel, iSrcStride, iHeight+1, iWidth+1, iDx<<1, iDy<<1, 4 );
    }
  }
}


// quarter-pixel upsampling for motion estimation
Void TComPredFilterMOMS::extMOMSUpSamplingQ ( TComPattern* pcPattern, TComMv cMvHalf, Pel* piDst, Int iDstStride )
{
  Int   iDx, iDy;

  Int   iWidth      = pcPattern->getROIYWidth();
  Int   iHeight     = pcPattern->getROIYHeight();

  Pel*  piSrc       = pcPattern->getROIY();
  Int   iSrcStride  = pcPattern->getPatternLStride();

  for ( iDy = -1; iDy <= 1; iDy++ )
  {
    for ( iDx = -1; iDx <= 1; iDx++ )
    {
      Int iDstDispX = ( cMvHalf.getHor() << 1 ) + iDx;
      Int iDstDispY = ( cMvHalf.getVer() << 1 ) + iDy;

      Int iSrcDispX = ( iDstDispX < 0 ? -1 : 0 );
      Int iSrcDispY = ( iDstDispY < 0 ? -1 : 0 );

      Int iPolyX    = ( iDstDispX < 0 ? 4+iDstDispX : iDstDispX );
      Int iPolyY    = ( iDstDispY < 0 ? 4+iDstDispY : iDstDispY );

      m_pcInterpIf->interpolate( piDst+(iDstDispX+iDstDispY*iDstStride), iDstStride, piSrc+(iSrcDispX+iSrcDispY*iSrcStride), iSrcStride, iHeight, iWidth, iPolyX<<1, iPolyY<<1, 4 );
    }
  }
}

// table for 1/8th sample interpolation
const Pel TComPredFilter4TapMOMS::sm_cFilterTable[8][4]=
{
  { 1820,        4552,        1820,           0, },
  {    0,           0,           0,           0, },
  {   32,         134,          86,           4, },
  {    0,           0,           0,           0, },
  {   14,         114,         114,          14, },
  {    0,           0,           0,           0, },
  {    4,          86,         134,          32, },
  {    0,           0,           0,           0, },
};


Void TComPredFilter4TapMOMS::xFilterHorInt ( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - iSrcStride - 1;

  const Int f0           = sm_cFilterTable[0][0];
  const Int f1           = sm_cFilterTable[0][1];

  for ( Int iY = 0; iY < iDstYMax + 3; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

#if HHI_MOMS4_NOMULT
      Int a0 = piSrc[0] + piSrc[2];
      Int a1 = piSrc[1] << 1;
      Int iTmp = a1 - a0;
      iSum  = iTmp << 2;
      iSum -= iTmp << 5;
      iSum += iTmp << 8;
      iSum += ( a0 + a1 ) << 11;
#else
      iSum  = f0 * ( piSrc[0] + piSrc[2] );
      iSum += f1 * piSrc[1];
#endif

      piTmp[iX] = ( iSum + (1<<4) ) >> 5;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterHorQ0( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDx )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - iSrcStride - 1;

  const Int f0           = sm_cFilterTable[iDx][0];
  const Int f1           = sm_cFilterTable[iDx][1];
  const Int f2           = sm_cFilterTable[iDx][2];
  const Int f3           = sm_cFilterTable[iDx][3];

  for ( Int iY = 0; iY < iDstYMax + 3; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

#if HHI_MOMS4_NOMULT
      Int iTmp = piSrc[1] + piSrc[2];
      iSum  = -(iTmp << 1);
      iSum += piSrc[3] << 2;
      iSum += ( piSrc[1] - piSrc[2] ) << 3;
      iSum += ( piSrc[0] - piSrc[2] ) << 5;
      iSum += iTmp << 7;
#else
      iSum  = f0 * piSrc[0];
      iSum += f1 * piSrc[1];
      iSum += f2 * piSrc[2];
      iSum += f3 * piSrc[3];
#endif

      piTmp[iX] = iSum;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterHorH( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDx )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - iSrcStride - 1;

  const Int f0           = sm_cFilterTable[iDx][0];
  const Int f1           = sm_cFilterTable[iDx][1];

  for ( Int iY = 0; iY < iDstYMax + 3; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

      Int a0 = ( piSrc[0] + piSrc[3] );
      Int a1 = ( piSrc[1] + piSrc[2] );

#if HHI_MOMS4_NOMULT
      Int iTmp = a1 - a0;
      iSum  = iTmp << 1;
      iSum -= iTmp << 4;
      iSum += a1 << 7;
#else
      iSum  = f0 * a0;
      iSum += f1 * a1;
#endif

      piTmp[iX] = iSum;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterHorQ1( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDx )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - iSrcStride - 1;

  const Int f0           = sm_cFilterTable[iDx][0];
  const Int f1           = sm_cFilterTable[iDx][1];
  const Int f2           = sm_cFilterTable[iDx][2];
  const Int f3           = sm_cFilterTable[iDx][3];

  for ( Int iY = 0; iY < iDstYMax + 3; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

      iSum  = f0 * piSrc[0];
      iSum += f1 * piSrc[1];
      iSum += f2 * piSrc[2];
      iSum += f3 * piSrc[3];

      piTmp[iX] = iSum;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterVerInt ( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[0][0];
  const Int f1           = sm_cFilterTable[0][1];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      Int a0 = ( *piTmp + 16 ) >> 5; piTmp += iTmpStride;
      Int a1 = ( *piTmp + 16 ) >> 5; piTmp += iTmpStride;
      Int a2 = ( *piTmp + 16 ) >> 5;

      iSum  = f0 * ( a0 + a2 );
      iSum += f1 * a1;
      iSum *= 9;

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterVerQ0( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDy )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[iDy][0];
  const Int f1           = sm_cFilterTable[iDy][1];
  const Int f2           = sm_cFilterTable[iDy][2];
  const Int f3           = sm_cFilterTable[iDy][3];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      iSum  = f0 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f1 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f2 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f3 * ( *piTmp );
      iSum *= 9;

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterVerH( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDy )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[iDy][0];
  const Int f1           = sm_cFilterTable[iDy][1];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      Int a0 = ( piTmp[0*iTmpStride] + piTmp[3*iTmpStride] );
      Int a1 = ( piTmp[1*iTmpStride] + piTmp[2*iTmpStride] );

      iSum  = f0 * a0;
      iSum += f1 * a1;
      iSum *= 9;

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

Void TComPredFilter4TapMOMS::xFilterVerQ1( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDy )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[iDy][0];
  const Int f1           = sm_cFilterTable[iDy][1];
  const Int f2           = sm_cFilterTable[iDy][2];
  const Int f3           = sm_cFilterTable[iDy][3];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      iSum  = f0 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f1 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f2 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f3 * ( *piTmp );
      iSum *= 9;

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

// main interpolation function
Void TComPredFilter4TapMOMS::interpolate( Pel* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDx, Int iDy, Int iDstStep, Int* piExtY )
{
  // temp buffer
  Int* pcTmpBufOrg;

  if ( piExtY == NULL ) 
    pcTmpBufOrg = new Int[iDstXMax*(iDstYMax+3)];
  else 
    pcTmpBufOrg = piExtY;

  // interpolation in horizontal direction
  switch( iDx )
  {
  case 0:
    xFilterHorInt ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep );
    break;
  case 2:
    xFilterHorQ0  ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep, iDx );
    break;
  case 4:
    xFilterHorH   ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep, iDx );
    break;
  case 6:
    xFilterHorQ1  ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep, iDx );
    break;
  }

  // interpolation in vertical direction
  switch ( iDy )
  {
  case 0:
    xFilterVerInt ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep );
    break;
  case 2:
    xFilterVerQ0  ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep, iDy );
    break;
  case 4:
    xFilterVerH   ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep, iDy );
    break;
  case 6:
    xFilterVerQ1  ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep, iDy );
    break;
  }

  if ( piExtY == NULL ) 
    delete[] pcTmpBufOrg;
}

// table for 1/4th sample interpolation
const Pel TComPredFilter6TapMOMS::sm_cFilterTable[8][6]=
{
  {  0, 8402,  18695,  8402,     0,     0, },
  {  0,     0,     0,     0,     0,     0, },
  { -7,   156,   560,   377,    26,    -3, },
  {  0,     0,     0,     0,     0,     0, },
  { -6,    77,   484,   484,    77,    -6, },
  {  0,     0,     0,     0,     0,     0, },
  { -3,    26,   377,   560,   156,    -7, },
  {  0,     0,     0,     0,     0,     0, },
};

Void TComPredFilter6TapMOMS::xFilterHorInt ( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - 2*iSrcStride - 2;

  const Int f1           = sm_cFilterTable[0][1];
  const Int f2           = sm_cFilterTable[0][2];

  for ( Int iY = 0; iY < iDstYMax + 5; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

      Int a0 = ( piSrc[1] + piSrc[3] );
      Int a1 = piSrc[2];

#if HHI_MOMS6_NOMULT
      iSum  = -a1;
      iSum += a0 << 1;
      iSum += a1 << 3;
      iSum += a0 << 4;
      iSum -= a0 << 6;
      iSum += ( a0 + a1 ) << 8;
      iSum += a1 << 11;
      iSum += a0 << 13;
      iSum += a1 << 14;
#else
      iSum  = f1 * a0;
      iSum += f2 * a1;
#endif

      piTmp[iX] = ( iSum + (1<<4) ) >> 5;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterHorQ0( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDx )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - 2*iSrcStride - 2;

  const Int f0           = sm_cFilterTable[iDx][0];
  const Int f1           = sm_cFilterTable[iDx][1];
  const Int f2           = sm_cFilterTable[iDx][2];
  const Int f3           = sm_cFilterTable[iDx][3];
  const Int f4           = sm_cFilterTable[iDx][4];
  const Int f5           = sm_cFilterTable[iDx][5];

  for ( Int iY = 0; iY < iDstYMax + 5; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

#if HHI_MOMS6_NOMULT
      Int iTemp1 = piSrc[0] + piSrc[3] + piSrc[5];
      iSum  = -iTemp1;
      iSum -= ( ( iTemp1 + piSrc[4] ) << 1 );
      iSum -= ( piSrc[0] + piSrc[1] + piSrc[3] + piSrc[4] ) << 2;
      iSum += ( piSrc[2] << 4 );
      iSum += ( piSrc[1] + piSrc[2] + piSrc[4] ) << 5;
      iSum += ( piSrc[1] + piSrc[3] ) << 7;
      iSum += ( piSrc[3] << 8 );
      iSum += ( piSrc[2] << 9 );
#else
      iSum  = f0 * piSrc[0];
      iSum += f1 * piSrc[1];
      iSum += f2 * piSrc[2];
      iSum += f3 * piSrc[3];
      iSum += f4 * piSrc[4];
      iSum += f5 * piSrc[5];
#endif

      piTmp[iX] = iSum;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterHorH( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDx )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - 2*iSrcStride - 2;

  const Int f0           = sm_cFilterTable[iDx][0];
  const Int f1           = sm_cFilterTable[iDx][1];
  const Int f2           = sm_cFilterTable[iDx][2];

  for ( Int iY = 0; iY < iDstYMax + 5; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

      Int a0 = ( piSrc[0] + piSrc[5] );
      Int a1 = ( piSrc[1] + piSrc[4] );
      Int a2 = ( piSrc[2] + piSrc[3] );

#if HHI_MOMS6_NOMULT
      a2  <<= 2;
      iSum  = a2 - a1;
      iSum += ( a0 - a1 ) << 1;
      iSum -= ( a0 + a2 ) << 3;
      iSum += a1 << 4;
      iSum += a1 << 6;
      iSum += a2 << 7;
#else
      iSum  = f0 * a0;
      iSum += f1 * a1;
      iSum += f2 * a2;
#endif

      piTmp[iX] = iSum;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterHorQ1( Int* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDx )
{
  Int* piTmp             = pcDst;
  const Int iTmpStride   = iDstXMax;
  const Pel* piSrcLn     = pcSrc - 2*iSrcStride - 2;

  const Int f0           = sm_cFilterTable[iDx][0];
  const Int f1           = sm_cFilterTable[iDx][1];
  const Int f2           = sm_cFilterTable[iDx][2];
  const Int f3           = sm_cFilterTable[iDx][3];
  const Int f4           = sm_cFilterTable[iDx][4];
  const Int f5           = sm_cFilterTable[iDx][5];

  for ( Int iY = 0; iY < iDstYMax + 5; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Pel* piSrc = piSrcLn + iX;

      iSum  = f0 * piSrc[0];
      iSum += f1 * piSrc[1];
      iSum += f2 * piSrc[2];
      iSum += f3 * piSrc[3];
      iSum += f4 * piSrc[4];
      iSum += f5 * piSrc[5];

      piTmp[iX] = iSum;
    }
    piTmp += iTmpStride;
    piSrcLn += iSrcStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterVerInt ( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc + iTmpStride;

  const Int f1           = sm_cFilterTable[0][1];
  const Int f2           = sm_cFilterTable[0][2];
  const Int f3           = sm_cFilterTable[0][3];
  const Int iF           = (13+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      Int a1 = ( *piTmp + 128 ) >> 8; piTmp += iTmpStride;
      Int a2 = ( *piTmp + 128 ) >> 8; piTmp += iTmpStride;
      Int a3 = ( *piTmp + 128 ) >> 8;

      iSum  = f1 * ( a1 + a3 );
      iSum += f2 * a2;

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterVerQ0( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDy )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[iDy][0];
  const Int f1           = sm_cFilterTable[iDy][1];
  const Int f2           = sm_cFilterTable[iDy][2];
  const Int f3           = sm_cFilterTable[iDy][3];
  const Int f4           = sm_cFilterTable[iDy][4];
  const Int f5           = sm_cFilterTable[iDy][5];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      iSum  = f0 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f1 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f2 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f3 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f4 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f5 * ( *piTmp );

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterVerH( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDy )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[iDy][0];
  const Int f1           = sm_cFilterTable[iDy][1];
  const Int f2           = sm_cFilterTable[iDy][2];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      Int a0 = ( piTmp[0*iTmpStride] + piTmp[5*iTmpStride] );
      Int a1 = ( piTmp[1*iTmpStride] + piTmp[4*iTmpStride] );
      Int a2 = ( piTmp[2*iTmpStride] + piTmp[3*iTmpStride] );

      iSum  = f0 * a0;
      iSum += f1 * a1;
      iSum += f2 * a2;

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

Void TComPredFilter6TapMOMS::xFilterVerQ1( Pel* pcDst, Int iDstStride, Int* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDstStep, Int iDy )
{
  Pel* piDst             = pcDst;
  const Int iTmpStride   = iSrcStride;
  const Int* piTmpLn     = pcSrc;

  const Int f0           = sm_cFilterTable[iDy][0];
  const Int f1           = sm_cFilterTable[iDy][1];
  const Int f2           = sm_cFilterTable[iDy][2];
  const Int f3           = sm_cFilterTable[iDy][3];
  const Int f4           = sm_cFilterTable[iDy][4];
  const Int f5           = sm_cFilterTable[iDy][5];
  const Int iF           = (16+6-g_uiBitIncrement);

  for ( Int iY = 0; iY < iDstYMax; iY++ )
  {
    for ( Int iX = 0; iX < iDstXMax; iX++ )
    {
      Int iSum;
      const Int* piTmp   = piTmpLn + iX;

      iSum  = f0 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f1 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f2 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f3 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f4 * ( *piTmp ); piTmp += iTmpStride;
      iSum += f5 * ( *piTmp );

      piDst[iDstStep*iX] = Clip( Pel( ( iSum + (1<<(iF-1)) ) >> iF ) );
    }
    piDst   += iDstStep*iDstStride;
    piTmpLn += iTmpStride;
  }
}

// main interpolation function
Void TComPredFilter6TapMOMS::interpolate( Pel* pcDst, Int iDstStride, Pel* pcSrc, Int iSrcStride, Int iDstYMax, Int iDstXMax, Int iDx, Int iDy, Int iDstStep, Int* piExtY )
{
  // temp buffer
  Int* pcTmpBufOrg;

  if ( piExtY == NULL ) 
    pcTmpBufOrg = new Int[iDstXMax*(iDstYMax+5)];
  else 
    pcTmpBufOrg = piExtY;

  // interpolation in horizontal direction
  switch( iDx )
  {
  case 0:
    xFilterHorInt ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep );
    break;
  case 2:
    xFilterHorQ0  ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep, iDx );
    break;
  case 4:
    xFilterHorH   ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep, iDx );
    break;
  case 6:
    xFilterHorQ1  ( pcTmpBufOrg, iDstXMax, pcSrc, iSrcStride, iDstYMax, iDstXMax, iDstStep, iDx );
    break;
  }

  // interpolation in vertical direction
  switch ( iDy )
  {
  case 0:
    xFilterVerInt ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep );
    break;
  case 2:
    xFilterVerQ0  ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep, iDy );
    break;
  case 4:
    xFilterVerH   ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep, iDy );
    break;
  case 6:
    xFilterVerQ1  ( pcDst, iDstStride, pcTmpBufOrg, iDstXMax, iDstYMax, iDstXMax, iDstStep, iDy );
    break;
  }

  if ( piExtY == NULL ) 
    delete[] pcTmpBufOrg;
}



// convert image samples to expansion coefficients
Void TComExpCoeff4TapMOMS::calcExpCoeffs( TComPicYuv *pcSrc, TComPicYuv *pcDst )
{
  xCalcCoeffsPlane( pcSrc->getLumaAddr(), pcDst->getLumaAddr(), pcDst->getHeight(),      pcDst->getWidth(),       pcDst->getStride() );
}


// internal function for initialization
Pel TComExpCoeff4TapMOMS::xInitRowCausalCoeff( const Pel* pcSrc,  Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iXMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << (m_iShift_IBDI+15) );

  for ( Int iX = 1; iX < iXMax; iX++ )
  {
    iSum      += Int( i16Q15Z_n ) * Int( pcSrc[iX] << m_iShift_IBDI );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// internal function for initialization
Pel TComExpCoeff4TapMOMS::xInitRowAntiCausalCoeff( const Pel* pcSrc,  Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iXMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << (m_iShift_IBDI+15) );

  for ( Int iX = 1; iX < iXMax; iX++ )
  {
    iSum      += Int( i16Q15Z_n ) * Int( pcSrc[-iX] << m_iShift_IBDI );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// internal function for initialization
Pel TComExpCoeff4TapMOMS::xInitColCausalCoeff( const Pel* pcSrc, Int iStride, Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iYMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << 15 );

  for ( Int iY = 1; iY < iYMax; iY++ )
  {
    pcSrc     += iStride;
    iSum      += Int( i16Q15Z_n ) * Int ( *pcSrc );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// internal function for initialization
Pel TComExpCoeff4TapMOMS::xInitColAntiCausalCoeff( const Pel* pcSrc, Int iStride, Int N )
{
	Pel i16Q15Z_n = m_i16Q15Z1;

	// only sum up 16 values
	const Int iYMax   = (N > 16) ? 16 : N;
	Int iSum          = Int( pcSrc[0] << 15 );

	for ( Int iY = 1; iY < iYMax; iY++ )
	{
		pcSrc     -= iStride;
		iSum      += Int( i16Q15Z_n ) * Int ( *pcSrc );
		i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
	}

	Pel cSum = (iSum + (1<<14)) >> 15;
	return cSum;
}

// convert image samples to expansion coefficients
Void TComExpCoeff4TapMOMS::xCalcCoeffsPlane( Pel* pcRefPic, Pel* pcCoeffs, Int iYMax, Int iXMax, Int iStride )
{
  Int iY, iX;

  Pel* pcSrc = pcRefPic;
  Pel* cBuff = new Pel[iYMax*iStride];
  Pel* pcCff = cBuff;

  // filter along rows
  for (iY=0; iY<iYMax; iY++)
  {
    // init causal coeff
    pcCff[0] = xInitRowCausalCoeff( pcSrc, iXMax );

    // causal filtering
    Pel cTmp = *pcCff;
    for (iX=1; iX<iXMax; iX++)
    {
      cTmp = Pel( pcSrc[iX] << m_iShift_IBDI ) - ( cTmp >> 1 );
      pcCff[iX] = cTmp;
    }

    // init anti-causal coeff
    cTmp = xInitRowAntiCausalCoeff( &pcSrc[iXMax-1], iXMax );
    pcCff[iXMax-1] = ( pcCff[iXMax-1] + cTmp - ( pcSrc[iXMax-1] << m_iShift_IBDI ) );

    // anti-causal filtering
    for (iX=iXMax-2; iX>=0; iX--)
    {
      cTmp = Pel( pcSrc[iX] << m_iShift_IBDI ) - ( cTmp >> 1 );
      pcCff[iX] = ( pcCff[iX] + cTmp - ( pcSrc[iX] << m_iShift_IBDI ) );
    }
    pcSrc += iStride;
    pcCff += iStride;
  }

  // filter along columns
  for (iX=0; iX<iXMax; iX++)
  {
    Pel cTmp;
    Pel* pcSrc = cBuff    + iX;
    Pel* pcCff = pcCoeffs + iX;

    // init causal coeff
    cTmp = xInitColCausalCoeff( pcSrc, iStride, iYMax );
    *pcCff = cTmp;

    // causal filtering
    for (iY=1; iY<iYMax; iY++)
    {
      pcSrc  += iStride;
      pcCff  += iStride;
      cTmp    = *pcSrc - ( cTmp >> 1 );
      *pcCff  = cTmp;
    }

    // init anti-causal coeff
    cTmp   = xInitColAntiCausalCoeff( pcSrc, iStride, iYMax );
    *pcCff = ( *pcCff + cTmp - *pcSrc );

    // anti-causal filtering
    for (iY=iYMax-2; iY>=0; iY--)
    {
      pcSrc -= iStride;
      pcCff -= iStride;
      cTmp   = *pcSrc - ( cTmp >> 1 );
      *pcCff = ( *pcCff + cTmp - *pcSrc );
    }
  }

  delete[] cBuff;
}

// convert image samples to expansion coefficients
Void TComExpCoeff6TapMOMS::calcExpCoeffs( TComPicYuv *pcSrc, TComPicYuv *pcDst )
{
  xCalcCoeffsPlane( pcSrc->getLumaAddr(), pcDst->getLumaAddr(), pcDst->getHeight(),      pcDst->getWidth(),       pcDst->getStride() );
}


// internal function for initialization
Pel TComExpCoeff6TapMOMS::xInitRowCausalCoeff( const Pel* pcSrc,  Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iXMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << (m_iShift_IBDI+15) );

  for ( Int iX = 1; iX < iXMax; iX++ )
  {
    iSum      += Int( i16Q15Z_n ) * Int( pcSrc[iX] << m_iShift_IBDI );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// internal function for initialization
Pel TComExpCoeff6TapMOMS::xInitRowAntiCausalCoeff( const Pel* pcSrc,  Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iXMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << (m_iShift_IBDI+15) );

  for ( Int iX = 1; iX < iXMax; iX++ )
  {
    iSum      += Int( i16Q15Z_n ) * Int( pcSrc[-iX] << m_iShift_IBDI );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// internal function for initialization
Pel TComExpCoeff6TapMOMS::xInitColCausalCoeff( const Pel* pcSrc, Int iStride, Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iYMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << 15 );

  for ( Int iY = 1; iY < iYMax; iY++ )
  {
    pcSrc     += iStride;
    iSum      += Int( i16Q15Z_n ) * Int ( *pcSrc );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// internal function for initialization
Pel TComExpCoeff6TapMOMS::xInitColAntiCausalCoeff( const Pel* pcSrc, Int iStride, Int N )
{
  Pel i16Q15Z_n = m_i16Q15Z1;

  // only sum up 16 values
  const Int iYMax   = (N > 16) ? 16 : N;
  Int iSum          = Int( pcSrc[0] << 15 );

  for ( Int iY = 1; iY < iYMax; iY++ )
  {
    pcSrc     -= iStride;
    iSum      += Int( i16Q15Z_n ) * Int ( *pcSrc );
    i16Q15Z_n  = xMult16x16    ( i16Q15Z_n, m_i16Q15Z1 );
  }

  Pel cSum = (iSum + (1<<14)) >> 15;
  return cSum;
}

// convert image samples to expansion coefficients
Void TComExpCoeff6TapMOMS::xCalcCoeffsPlane( Pel* pcRefPic, Pel* pcCoeffs, Int iYMax, Int iXMax, Int iStride )
{
  Int iY, iX;

  Pel* pcSrc = pcRefPic;
  Pel* cBuff = new Pel[iYMax*iStride];
  Pel* pcCff = cBuff;

  // filter along rows
  for (iY=0; iY<iYMax; iY++)
  {
    // init causal coeff
    pcCff[0] = xInitRowCausalCoeff( pcSrc, iXMax );

    // causal filtering
    Pel cTmp = *pcCff;
    for (iX=1; iX<iXMax; iX++)
    {
      cTmp = ( pcSrc[iX] << m_iShift_IBDI ) - ( ( 5 * cTmp + 4 ) >> 3 );
      pcCff[iX] = cTmp;
    }

    // init anti-causal coeff
    cTmp = xInitRowAntiCausalCoeff( &pcSrc[iXMax-1], iXMax );
    pcCff[iXMax-1] = ( pcCff[iXMax-1] + cTmp - ( pcSrc[iXMax-1] << m_iShift_IBDI ) );

    // anti-causal filtering
    for (iX=iXMax-2; iX>=0; iX--)
    {
      cTmp = ( pcSrc[iX] << m_iShift_IBDI ) - ( ( 5 * cTmp + 4 ) >> 3 );
      pcCff[iX] = ( pcCff[iX] + cTmp - ( pcSrc[iX] << m_iShift_IBDI ) );
    }
    pcSrc += iStride;
    pcCff += iStride;
  }

  // filter along columns
  for (iX=0; iX<iXMax; iX++)
  {
    Pel cTmp;
    Pel* pcSrc = cBuff    + iX;
    Pel* pcCff = pcCoeffs + iX;

    // init causal coeff
    cTmp = xInitColCausalCoeff( pcSrc, iStride, iYMax );
    *pcCff = cTmp;

    // causal filtering
    for (iY=1; iY<iYMax; iY++)
    {
      pcSrc  += iStride;
      pcCff  += iStride;
      cTmp    = *pcSrc - ( ( 5 * cTmp + 4 ) >> 3 );
      *pcCff  = cTmp;
    }

    // init anti-causal coeff
    cTmp   = xInitColAntiCausalCoeff( pcSrc, iStride, iYMax );
    *pcCff = ( *pcCff + cTmp - *pcSrc );

    // anti-causal filtering
    for (iY=iYMax-2; iY>=0; iY--)
    {
      pcSrc -= iStride;
      pcCff -= iStride;
      cTmp   = *pcSrc - ( ( 5 * cTmp + 4 ) >> 3 );
      *pcCff = ( *pcCff + cTmp - *pcSrc );
    }
  }

  delete[] cBuff;
}

#endif /* HHI_INTERP_FILTER */
